/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "zconf.tab.h"
#include "zconf.h"

const char conf_def_filename[] = ".config";
char conf_filename[128];

int conf_read(const char *name)
{
	FILE *in;
	char line[128];
	char *p, *p2;
	int lineno = 0;
	struct symbol *sym;
	struct statement *stmt;
	struct expr *e;
	int i;

	if (name) {
		in = fopen(name, "r");
		if (in)
			strcpy(conf_filename, name);
	} else {
		strcpy(conf_filename, conf_def_filename);
		in = fopen(conf_filename, "r");
		if (!in) {
			sym = lookup_symbol("ARCH", 0);
			sym_calc_value(sym);
			name = sym_get_string_value(sym);
			sprintf(conf_filename, "arch/%s/defconfig", name);
			in = fopen(conf_filename, "r");
		}
		if (in)
			printf("#\n"
			       "# using defaults found in %s\n"
			       "#\n", conf_filename);
		name = conf_filename;
	}

	if (!in)
		return 1;

	for_all_symbols(i, sym) {
		sym->flags |= SYMBOL_NEW;
		switch (sym->type) {
		case SYMBOL_INT:
		case SYMBOL_HEX:
		case SYMBOL_STRING:
			if (S_VAL(sym->def)) {
				free(S_VAL(sym->def));
				S_VAL(sym->def) = NULL;
			}
		}
	}

	while (fgets(line, 128, in)) {
		lineno++;
		switch (line[0]) {
		case '#':
			if (memcmp(line + 2, "CONFIG_", 7))
				continue;
			p = strchr(line + 9, ' ');
			if (!p)
				continue;
			*p++ = 0;
			if (strncmp(p, "is not set", 10))
				continue;
			//printf("%s -> n\n", line + 9);
			sym = lookup_symbol(line + 9, 0);
			switch (sym->type) {
			case SYMBOL_BOOLEAN:
			case SYMBOL_TRISTATE:
				sym->def = symbol_no.curr;
				sym->flags &= ~SYMBOL_NEW;
				break;
			}
			break;
		case 'C':
			if (memcmp(line, "CONFIG_", 7))
				continue;
			p = strchr(line + 7, '=');
			if (!p)
				continue;
			*p++ = 0;
			p2 = strchr(p, '\n');
			if (p2)
				*p2 = 0;
			//printf("%s -> %s\n", line + 7, p);
			sym = lookup_symbol(line + 7, 0);
			switch (sym->type) {
			case SYMBOL_BOOLEAN:
				sym->def = symbol_yes.curr;
				sym->flags &= ~SYMBOL_NEW;
				break;
			case SYMBOL_TRISTATE:
				if (p[0] == 'm')
					sym->def = symbol_mod.curr;
				else
					sym->def = symbol_yes.curr;
				sym->flags &= ~SYMBOL_NEW;
				break;
			case SYMBOL_STRING:
				if (p[0] == '"' && p[strlen(p)-1] == '"') {
					p++;
					p[strlen(p)-1] = 0;
					for (p2 = p; (p2 = strchr(p2, '\\')); p2++) {
						if (p2[1] == '\\' || p2[1] == '"')
							memmove(p2, p2 + 1, strlen(p2));
					}
				}
			case SYMBOL_INT:
			case SYMBOL_HEX:
				if (sym_string_valid(sym, p)) {
					S_VAL(sym->def) = strdup(p);
					sym->flags &= ~SYMBOL_NEW;
				} else
					fprintf(stderr, "%s:%d:symbol value '%s' invalid for %s\n", name, lineno, p, sym->name);
				break;
			}
			if (sym_is_choice_value(sym)) {
				stmt = sym_get_choice_stmt(sym);
				switch (S_TRI(sym->def)) {
				case mod:
					if (S_TRI(stmt->def->def) == yes)
						/* warn? */;
					break;
				case yes:
					if (S_TRI(stmt->def->def) != no)
						/* warn? */;
					S_VAL(stmt->def->def) = sym;
					break;
				case no:
					break;
				}
				S_TRI(stmt->def->def) = S_TRI(sym->def);
			}
			break;
		case '\n':
			break;
		default:
			continue;
		}
	}
	fclose(in);

	for_all_symbols(i, sym) {
		if (!sym_is_choice(sym))
			continue;
		stmt = sym_get_choice_stmt(sym);
		sym->flags &= ~SYMBOL_NEW;
		for (e = stmt->dep; e; e = e->left.expr)
			sym->flags |= e->right.sym->flags & SYMBOL_NEW;
	}

	sym_change_count = 1;

	return 0;
}

int conf_write(const char *name)
{
	FILE *out, *out_h;
	struct symbol *sym;
	struct menu *menu;
	char oldname[128];
	int type, l;
	const char *str;

	out = fopen(".tmpconfig", "w");
	if (!out)
		return 1;
	out_h = fopen(".tmpconfig.h", "w");
	if (!out_h)
		return 1;
	fprintf(out, "#\n"
		     "# Automatically generated make config: don't edit\n"
		     "#\n");
	fprintf(out_h, "/*\n"
		       " * Automatically generated C config: don't edit\n"
		       " */\n"
		       "#define AUTOCONF_INCLUDED\n");

	if (!sym_change_count)
		sym_clear_all_valid();

	menu = rootmenu.list;
	while (menu) {
		sym = menu->sym;
		if (!sym) {
			if (!menu_is_visible(menu))
				goto next;
			str = menu_get_prompt(menu);
			fprintf(out, "\n"
				     "#\n"
				     "# %s\n"
				     "#\n", str);
			fprintf(out_h, "\n"
				       "/*\n"
				       " * %s\n"
				       " */\n", str);
		} else if (!(sym->flags & SYMBOL_CHOICE)) {
			sym_calc_value(sym);
			if (!(sym->flags & SYMBOL_WRITE))
				goto next;
			sym->flags &= ~SYMBOL_WRITE;
			type = sym->type;
			if (type == SYMBOL_TRISTATE) {
				sym_calc_value(modules_sym);
				if (S_TRI(modules_sym->curr) == no)
					type = SYMBOL_BOOLEAN;
			}
			switch (type) {
			case SYMBOL_BOOLEAN:
			case SYMBOL_TRISTATE:
				switch (sym_get_tristate_value(sym)) {
				case no:
					fprintf(out, "# CONFIG_%s is not set\n", sym->name);
					fprintf(out_h, "#undef CONFIG_%s\n", sym->name);
					break;
				case mod:
					fprintf(out, "CONFIG_%s=m\n", sym->name);
					fprintf(out_h, "#define CONFIG_%s_MODULE 1\n", sym->name);
					break;
				case yes:
					fprintf(out, "CONFIG_%s=y\n", sym->name);
					fprintf(out_h, "#define CONFIG_%s 1\n", sym->name);
					break;
				}
				break;
			case SYMBOL_STRING:
				// fix me
				str = sym_get_string_value(sym);
				fprintf(out, "CONFIG_%s=\"", sym->name);
				fprintf(out_h, "#define CONFIG_%s \"", sym->name);
				do {
					l = strcspn(str, "\"\\");
					if (l) {
						fwrite(str, l, 1, out);
						fwrite(str, l, 1, out_h);
					}
					str += l;
					while (*str == '\\' || *str == '"') {
						fprintf(out, "\\%c", *str);
						fprintf(out_h, "\\%c", *str);
						str++;
					}
				} while (*str);
				fputs("\"\n", out);
				fputs("\"\n", out_h);
				break;
			case SYMBOL_HEX:
				str = sym_get_string_value(sym);
				if (str[0] != '0' || (str[1] != 'x' && str[1] != 'X')) {
					fprintf(out, "CONFIG_%s=%s\n", sym->name, str);
					fprintf(out_h, "#define CONFIG_%s 0x%s\n", sym->name, str);
					break;
				}
			case SYMBOL_INT:
				str = sym_get_string_value(sym);
				fprintf(out, "CONFIG_%s=%s\n", sym->name, str);
				fprintf(out_h, "#define CONFIG_%s %s\n", sym->name, str);
				break;
			}
		}

	next:
		if (menu->list) {
			menu = menu->list;
			continue;
		}
		if (menu->next)
			menu = menu->next;
		else while ((menu = menu->parent)) {
			if (menu->next) {
				menu = menu->next;
				break;
			}
		}
	}
	fclose(out);
	fclose(out_h);

	if (!name) {
		rename(".tmpconfig.h", "include/linux/autoconf.h");
		name = conf_def_filename;
		file_write_dep(NULL);
	} else
		unlink(".tmpconfig.h");

	sprintf(oldname, "%s.old", name);
	rename(name, oldname);
	if (rename(".tmpconfig", name))
		return 1;
	strcpy(conf_filename, name);

	sym_change_count = 0;

	return 0;
}
