/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#ifndef EXPR_H
#define EXPR_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#ifndef __cplusplus
#include <stdbool.h>
#endif

struct file {
	struct file *next;
	struct file *parent;
	struct statement *stmt;
	struct statement *last_stmt;
	char *name;
	int lineno;
	int flags;
};

#define FILE_BUSY		0x0001
#define FILE_SCANNED		0x0002
#define FILE_PRINTED		0x0004

typedef enum tristate {
	no, mod, yes
} tristate;

struct expr {
	int token;
	union {
		struct expr *expr;
		struct symbol *sym;
	} left, right;
};

//#define E_NO		(0)
//#define E_MOD		(1)
//#define E_YES		(2)
#define E_TRI(ev)	((ev).tri)
#define E_EXPR(ev)	((ev).expr)
#define E_CALC(ev)	(E_TRI(ev) = expr_calc_value(E_EXPR(ev)))

#define E_OR(dep1, dep2)	(((dep1)>(dep2))?(dep1):(dep2))
#define E_AND(dep1, dep2)	(((dep1)<(dep2))?(dep1):(dep2))
#define E_NOT(dep)		(2-(dep))

struct expr_value {
	struct expr *expr;
	tristate tri;
};

#define S_VAL(sv)	((sv).value)
#define S_TRI(sv)	((sv).tri)
#define S_EQ(sv1, sv2)	(S_VAL(sv1) == S_VAL(sv2) || !strcmp(S_VAL(sv1), S_VAL(sv2)))

struct symbol_value {
	void *value;
	tristate tri;
};

struct symbol {
	struct symbol *next;
	char *name;
	char *help;
	int type;
	struct symbol_value curr, def;
	tristate visible;
	int flags;
	struct statement *stmt;
	struct expr *dep, *dep2;
	struct menu *menu;
};

#define for_all_symbols(i, sym) for (i = 0; i < 257; i++) for (sym = symbol_hash[i]; sym; sym = sym->next) if (sym->type != SYMBOL_OTHER)

#define SYMBOL_UNKNOWN		0
#define SYMBOL_BOOLEAN		1
#define SYMBOL_TRISTATE		2
#define SYMBOL_INT		3
#define SYMBOL_HEX		4
#define SYMBOL_STRING		5
#define SYMBOL_OTHER		6

#define SYMBOL_YES		0x0001
#define SYMBOL_MOD		0x0002
#define SYMBOL_NO		0x0004
#define SYMBOL_CONST		0x0007
#define SYMBOL_CHECK		0x0008
#define SYMBOL_CHOICE		0x0010
#define SYMBOL_CHOICEVAL	0x0020
#define SYMBOL_PRINTED		0x0040
#define SYMBOL_VALID		0x0080
#define SYMBOL_OPTIONAL		0x0100
#define SYMBOL_WRITE		0x0200
#define SYMBOL_CHANGED		0x0400
#define SYMBOL_NEW		0x0800

struct statement {
	struct statement *next;
	struct symbol *sym;
	int token;
	char *prompt;
	struct symbol *def;
	struct expr_value visible;
	struct expr *dep;
	struct expr *dep2;
	struct menu *menu;
	struct file *file;
	struct statement *next_pos;
	int lineno;
};

#define for_all_stmt_token(sym, st, tok) \
	for (st = sym->stmt; st; st = st->next) \
		if (st->token == (tok))
#define for_all_prompts(sym, st) for_all_stmt_token(sym, st, T_PROMPT)
#define for_all_defaults(sym, st) for_all_stmt_token(sym, st, T_DEFAULT)
#define for_all_choices(sym, st) for_all_stmt_token(sym, st, '^')

#define T_WORD		300
#define T_UNEQUAL	301
#define T_OR		302
#define T_AND		303
#define T_IF		304
#define T_ELSE		305
#define T_PROMPT	306
#define T_DEFAULT	307

struct menu {
	struct menu *next;
	struct menu *parent;
	struct menu *list;
	struct symbol *sym;
	struct statement *prompt;
	struct expr *dep;
	//char *help;
	struct file *file;
	int lineno;
	//void *data;
};

extern struct symbol *symbol_hash[257];
extern struct menu rootmenu;
extern struct file *file_list;
extern struct file *current_file;
struct file *lookup_file(const char *name);

extern struct symbol symbol_yes, symbol_no, symbol_mod;
extern struct symbol *modules_sym;
extern int cdebug;
extern int print_type;
struct expr *expr_alloc_symbol(struct symbol *sym);
struct expr *expr_alloc_one(int token, struct expr *ce);
struct expr *expr_alloc_two(int token, struct expr *e1, struct expr *e2);
struct expr *expr_alloc_comp(int token, struct symbol *s1, struct symbol *s2);
struct expr *expr_alloc_and(struct expr *e1, struct expr *e2);
struct expr *expr_copy(struct expr *org);
void expr_free(struct expr *e);
int expr_eq(struct expr *e1, struct expr *e2);
void expr_eliminate_eq(struct expr **ep1, struct expr **ep2);
tristate expr_calc_value(struct expr *e);
struct expr *expr_eliminate_yn(struct expr *e);
struct expr *expr_trans_bool(struct expr *e);
struct expr *expr_eliminate_dups(struct expr *e);
struct expr *expr_transform(struct expr *e);
int expr_contains_symbol(struct expr *dep, struct symbol *sym);
bool expr_depends_symbol(struct expr *dep, struct symbol *sym);
struct expr *expr_extract_eq_and(struct expr **ep1, struct expr **ep2);
struct expr *expr_extract_eq_or(struct expr **ep1, struct expr **ep2);
void expr_extract_eq(int token, struct expr **ep, struct expr **ep1, struct expr **ep2);
struct expr *expr_trans_compare(struct expr *e, int token, struct symbol *sym);

void fprint_expr(FILE *out, struct expr *e, int prevtoken);
void print_expr(int mask, struct expr *e, int prevtoken);

static inline int expr_is_yes(struct expr *e)
{
	return !e || (e->token == T_WORD && e->left.sym == &symbol_yes);
}

static inline int expr_is_no(struct expr *e)
{
	return e && (e->token == T_WORD && e->left.sym == &symbol_no);
}


#ifdef __cplusplus
}
#endif

#endif /* EXPR_H */
