/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <qapplication.h>
#include <qmainwindow.h>
#include <qtoolbar.h>
#include <qvbox.h>
#include <qsplitter.h>
#include <qlistview.h>
#include <qtextview.h>
#include <qlineedit.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qaction.h>
#include <qheader.h>
#include <qfiledialog.h>
#include <stdlib.h>
#include "zconf.h"
#include "zconf.tab.h"
#include "qconf.h"
#include "version.h"

#include "qconf.moc"
#include "images.c"

static QApplication *configApp;

/*
 * update all the children of a menu entry
 *   removes/adds the entries from the parent widget as necessary
 *
 * parent: either the menu list widget or a menu entry widget
 * menu: entry to be updated
 */
template <class P>
static void updateMenuList(P* parent, struct menu* menu, bool showFull)
{
	struct menu* child;
	ConfigList* list = parent->listView();
	ConfigItem* item;
	ConfigItem* last;
	bool visible;
	bool showAll = list->showAll;

#if 0
	printf("u: %p,%p,", parent, menu);
	if (!menu) {
		printf("<null>\n");
		return;
	} else if (menu->sym && menu->sym->name)
		printf("sym %s\n", menu->sym->name);
	else if (menu->sym && menu->prompt)
		printf("choice %s\n", menu->prompt->prompt);
	else if (menu->prompt)
		printf("prompt %s\n", menu->prompt->prompt);
	else
		printf("empty\n");
#else
	if (!menu)
		return;
#endif

	last = 0;
	for (child = menu->list; child; child = child->next) {
		visible = menu_is_visible(child);
		item = last ? last->nextSibling() : parent->firstChild();

		if ((showAll || visible) && showFull) {
			if (!item || item->menu != child)
				item = new ConfigItem(parent, last, child);
			item->setEnabled(visible);
			item->updateMenu();
			if (child->prompt && child->prompt->token == T_MENU)
				updateMenuList(item, child, list->showFull);
			else
				updateMenuList(item, child, true);
			last = item;
		} else {
			if (item && item->menu == child) {
				last = parent->firstChild();
				if (last == item)
					last = 0;
				else while (last->nextSibling() != item)
					last = last->nextSibling();
				delete item;
			}
		}
	}
}

#if QT_VERSION >= 300
/*
 * set the new data
 * TODO check the value
 */
void ConfigItem::okRename(int col)
{
	Parent::okRename(col);
	sym_set_string_value(menu->sym, text(dataColIdx).latin1());
}
#endif

/*
 * update the displayed of a menu entry
 */
void ConfigItem::updateMenu(void)
{
	ConfigList* list;
	struct symbol* sym;
	int type;
	tristate expr;
	bool update;

	list = listView();
	update = doInit;
	if (update)
		doInit = false;
	else
		update = list->updateAll;

	sym = menu->sym;
	if (!sym) {
		if (update) {
			setText(promptColIdx, menu_get_prompt(menu));
			if (menu->prompt && menu->prompt->token == T_MENU && !list->showFull)
				setPixmap(promptColIdx, list->menuPix);
			else
				setPixmap(promptColIdx, 0);
		}
		return;
	}

	sym_calc_value(sym);
	if (!(sym->flags & SYMBOL_CHANGED) && !update)
		return;

	sym->flags &= ~SYMBOL_CHANGED;

	setText(nameColIdx, menu->sym->name);

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		char ch;

		setText(promptColIdx, QString().sprintf("%s%s", menu_get_prompt(menu),
				sym_has_value(sym) ? "" : " (NEW)"));
		if (!sym_is_changable(sym) && !list->showAll) {
			setText(noColIdx, 0);
			setText(modColIdx, 0);
			setText(yesColIdx, 0);
			break;
		}
		expr = sym_get_tristate_value(sym);
		switch (expr) {
		case yes:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(promptColIdx, list->choiceYesPix);
			else
				setPixmap(promptColIdx, list->symbolYesPix);
			setText(yesColIdx, "Y");
			ch = 'Y';
			break;
		case mod:
			setPixmap(promptColIdx, list->symbolModPix);
			setText(modColIdx, "M");
			ch = 'M';
			break;
		default:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(promptColIdx, list->choiceNoPix);
			else
				setPixmap(promptColIdx, list->symbolNoPix);
			setText(noColIdx, "N");
			ch = 'N';
			break;
		}
		if (expr != no)
			setText(noColIdx, sym_tristate_within_range(sym, no) ? "_" : 0);
		if (expr != mod)
			setText(modColIdx, sym_tristate_within_range(sym, mod) ? "_" : 0);
		if (expr != yes)
			setText(yesColIdx, sym_tristate_within_range(sym, yes) ? "_" : 0);

#if 0
		if (type == SYMBOL_BOOLEAN)
			setText(promptColIdx, QString().sprintf("[%c] %s", ch, menu_get_prompt(menu)));
		else
			setText(promptColIdx, QString().sprintf("<%c> %s", ch, menu_get_prompt(menu)));
		setText(promptColIdx, menu_get_prompt(menu));
#endif
		setText(dataColIdx, QChar(ch));
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		const char* data;

		data = sym_get_string_value(sym);
#if QT_VERSION >= 300
		setRenameEnabled(list->mapIdx(dataColIdx), TRUE);
#endif
		setText(dataColIdx, data);
		if (type == SYMBOL_STRING)
			setText(promptColIdx, QString().sprintf("%s: %s", menu_get_prompt(menu), data));
		else
			setText(promptColIdx, QString().sprintf("(%s) %s", data, menu_get_prompt(menu)));
		break;
	}
}

/*
 * construct a menu entry
 */
void ConfigItem::init(void)
{
	ConfigList* list = listView();
#if QT_VERSION < 300
	setEnabled(TRUE);
#endif
	//menu->data = this;
	if (!list->showFull)
		setOpen(TRUE);
	doInit= true;
}

/*
 * destruct a menu entry
 */
ConfigItem::~ConfigItem(void)
{
	//menu->data = 0;
}

void ConfigLineEdit::show(ConfigItem* i)
{
	item = i;
	if (sym_get_string_value(item->menu->sym))
		setText(sym_get_string_value(item->menu->sym));
	else
		setText(0);
	Parent::show();
	setFocus();
}

void ConfigLineEdit::keyPressEvent(QKeyEvent* e)
{
	switch (e->key()) {
	case Key_Escape:
		break;
	case Key_Return:
	case Key_Enter:
		sym_set_string_value(item->menu->sym, text().latin1());
		emit lineChanged(item);
		break;
	default:
		Parent::keyPressEvent(e);
		return;
	}
	e->accept();
	hide();
}

ConfigList::ConfigList(QWidget* p)
	: Parent(p),
	  updateAll(false),
	  symbolYesPix(xpm_symbol_yes), symbolModPix(xpm_symbol_mod), symbolNoPix(xpm_symbol_no),
	  choiceYesPix(xpm_choice_yes), choiceNoPix(xpm_choice_no), menuPix(xpm_menu), menuInvPix(xpm_menu_inv),
	  showAll(false), showName(false), showRange(false), showData(false), showFull(false),
	  rootEntry(0)
{
	int i;

	setSorting(-1);
	setRootIsDecorated(TRUE);

	for (i = 0; i < colNr; i++)
		colMap[i] = colRevMap[i] = -1;
	addColumn(promptColIdx, "Option");

	reinit();
}

void ConfigList::reinit(void)
{
	removeColumn(dataColIdx);
	removeColumn(yesColIdx);
	removeColumn(modColIdx);
	removeColumn(noColIdx);
	removeColumn(nameColIdx);

	if (showName)
		addColumn(nameColIdx, "Name");
	if (showRange) {
		addColumn(noColIdx, "N");
		addColumn(modColIdx, "M");
		addColumn(yesColIdx, "Y");
	}
	if (showData)
		addColumn(dataColIdx, "Value");

	updateListAll();
}

void ConfigList::updateList(ConfigItem* item)
{
	(void)item;	// unused so far
	updateMenuList(this, rootEntry, true);
}

void ConfigList::setValue(ConfigItem* item, tristate val)
{
	struct symbol* sym;
	int type;
	tristate oldval;

#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif
	sym = item->menu->sym;
	if (!sym)
		return;

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldval = sym_get_tristate_value(sym);

		if (!sym_set_tristate_value(sym, val))
			return;
		if (oldval == no && item->menu->list)
			item->setOpen(TRUE);
		emit symbolChanged(item);
		break;
	}
}

void ConfigList::changeValue(ConfigItem* item)
{
	struct symbol* sym;
	struct menu* menu;
	int type, oldexpr, newexpr;

#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif

	menu = item->menu;
	sym = menu->sym;
	if (!sym) {
		if (item->menu->list)
			item->setOpen(!item->isOpen());
		return;
	}

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldexpr = sym_get_tristate_value(sym);
		newexpr = sym_toggle_tristate_value(sym);
		if (item->menu->list) {
			if (oldexpr == newexpr)
				item->setOpen(!item->isOpen());
			else if (oldexpr == no)
				item->setOpen(TRUE);
		}
		if (oldexpr != newexpr)
			emit symbolChanged(item);
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
#if QT_VERSION >= 300
		if (colMap[dataColIdx] >= 0)
			item->startRename(colMap[dataColIdx]);
		else
#endif
			lineEdit->show(item);
		break;
	}
}

void ConfigList::setRootMenu(struct menu *menu)
{
	if (rootEntry == menu)
		return;
	if (menu && (!menu->prompt || menu->prompt->token != T_MENU))
		return;
	updateMenuList(this, rootEntry, false);
	rootEntry = menu;
	updateListAll();
}

void ConfigList::keyPressEvent(QKeyEvent* ev)
{
	QListViewItem* i = currentItem();
	ConfigItem* item;
	struct menu *menu;

	if (ev->key() == Key_Escape && !showFull) {
		struct menu *oldroot, *newroot;

		ev->accept();
		oldroot = rootEntry;
		newroot = menu_get_parent_menu(oldroot);
		if (newroot == oldroot)
			return;
		setRootMenu(newroot);

		QListViewItemIterator it(this);
		while ((item = (ConfigItem*)it.current())) {
			if (item->menu == oldroot) {
				setCurrentItem(item);
				ensureItemVisible(item);
				break;
			}
			it++;
		}
		return;
	}

	if (!i) {
		Parent::keyPressEvent(ev);
		return;
	}
	item = (ConfigItem*)i;

	switch (ev->key()) {
	case Key_Return:
	case Key_Enter:
		menu = item->menu;
		if (!showFull && menu->prompt && menu->prompt->token == T_MENU) {
			emit menuSelected(menu);
			break;
		}
	case Key_Space:
		changeValue(item);
		break;
	case Key_N:
		setValue(item, no);
		break;
	case Key_M:
		setValue(item, mod);
		break;
	case Key_Y:
		setValue(item, yes);
		break;
	default:
		Parent::keyPressEvent(ev);
		return;
	}
	ev->accept();
}

void ConfigList::contentsMousePressEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMousePressEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMousePressEvent(e);
}

void ConfigList::contentsMouseReleaseEvent(QMouseEvent* e)
{
	QPoint p(contentsToViewport(e->pos()));
	ConfigItem* item = (ConfigItem*)itemAt(p);
	const QPixmap* pm;
	int idx, x;

	if (!item)
		goto skip;

	x = header()->offset() + p.x();
	idx = colRevMap[header()->sectionAt(x)];
	switch (idx) {
	case promptColIdx:
		pm = item->pixmap(promptColIdx);
		if (pm) {
			int off = header()->sectionPos(0) + itemMargin() +
				treeStepSize() * (item->depth() + (rootIsDecorated() ? 1 : 0));
			if (x >= off && x < off + pm->width()) {
				if (item->menu->sym)
					changeValue(item);
				else
					emit menuSelected(item->menu);
			}
		}
		break;
	case noColIdx:
		setValue(item, no);
		break;
	case modColIdx:
		setValue(item, mod);
		break;
	case yesColIdx:
		setValue(item, yes);
		break;
	case dataColIdx:
		changeValue(item);
		break;
	}

	//printf("contentsMouseReleaseEvent: %d,%d\n", p.x(), p.y());
skip:
	QListView::contentsMouseReleaseEvent(e);
}

void ConfigList::contentsMouseMoveEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMouseMoveEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMouseMoveEvent(e);
}

void ConfigList::contentsMouseDoubleClickEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMouseDoubleClickEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMouseDoubleClickEvent(e);
}

/*
 * Construct the complete config widget
 */
ConfigView::ConfigView(void)
{
	QMenuBar* menu;
	QSplitter* split1;
	QSplitter* split2;

	split1 = new QSplitter(this);
	split1->setOrientation(QSplitter::Horizontal);
	setCentralWidget(split1);

	menuList = new ConfigList(split1);

	split2 = new QSplitter(split1);
	split2->setOrientation(QSplitter::Vertical);

	// create config tree
	QVBox* box = new QVBox(split2);
	configList = new ConfigList(box);
	configList->lineEdit = new ConfigLineEdit(box);
	configList->lineEdit->hide();
	configList->connect(configList, SIGNAL(symbolChanged(ConfigItem*)),
			    configList, SLOT(updateList(ConfigItem*)));
	configList->connect(configList, SIGNAL(symbolChanged(ConfigItem*)),
			    menuList, SLOT(updateList(ConfigItem*)));
	configList->connect(configList->lineEdit, SIGNAL(lineChanged(ConfigItem*)),
			    SLOT(updateList(ConfigItem*)));
	configList->connect(configList, SIGNAL(menuSelected(struct menu *)),
			    SLOT(setRootMenu(struct menu *)));
	configList->connect(menuList, SIGNAL(menuSelected(struct menu *)),
			    SLOT(setRootMenu(struct menu *)));

	helpText = new QTextView(split2);

	setTabOrder(configList, helpText);
	configList->setFocus();

	menu = menuBar();
	toolBar = new QToolBar("Tools", this);

	QAction *backAction = new QAction("Back", QPixmap(xpm_back), "Back", 0, this);
	  backAction->setEnabled(FALSE);
	QAction *quitAction = new QAction("Quit", "&Quit", CTRL+Key_Q, this);
	  connect(quitAction, SIGNAL(activated()), SLOT(close()));
	QAction *loadAction = new QAction("Load", QPixmap(xpm_load), "&Load", CTRL+Key_L, this);
	  connect(loadAction, SIGNAL(activated()), SLOT(loadConfig()));
	QAction *saveAction = new QAction("Save", QPixmap(xpm_save), "&Save", CTRL+Key_S, this);
	  connect(saveAction, SIGNAL(activated()), SLOT(saveConfig()));
	QAction *saveAsAction = new QAction("Save As...", "Save &As...", 0, this);
	  connect(saveAsAction, SIGNAL(activated()), SLOT(saveConfigAs()));
	QAction *singleViewAction = new QAction("Single View", QPixmap(xpm_single_view), "Split View", 0, this);
	  connect(singleViewAction, SIGNAL(activated()), SLOT(showSingleView()));
	QAction *splitViewAction = new QAction("Split View", QPixmap(xpm_split_view), "Split View", 0, this);
	  connect(splitViewAction, SIGNAL(activated()), SLOT(showSplitView()));
	QAction *fullViewAction = new QAction("Full View", QPixmap(xpm_tree_view), "Full View", 0, this);
	  connect(fullViewAction, SIGNAL(activated()), SLOT(showFullView()));

	QAction *showAllAction = new QAction(NULL, "Show All", 0, this);
	  showAllAction->setToggleAction(TRUE);
	  showAllAction->setOn(configList->showAll);
	  connect(showAllAction, SIGNAL(toggled(bool)), SLOT(setShowAll(bool)));
	QAction *showNameAction = new QAction(NULL, "Show Name", 0, this);
	  showNameAction->setToggleAction(TRUE);
	  showNameAction->setOn(configList->showName);
	  connect(showNameAction, SIGNAL(toggled(bool)), SLOT(setShowName(bool)));
	QAction *showRangeAction = new QAction(NULL, "Show Range", 0, this);
	  showRangeAction->setToggleAction(TRUE);
	  showRangeAction->setOn(configList->showRange);
	  connect(showRangeAction, SIGNAL(toggled(bool)), SLOT(setShowRange(bool)));
	QAction *showDataAction = new QAction(NULL, "Show Data", 0, this);
	  showDataAction->setToggleAction(TRUE);
	  showDataAction->setOn(configList->showData);
	  connect(showDataAction, SIGNAL(toggled(bool)), SLOT(setShowData(bool)));

	// init tool bar
	backAction->addTo(toolBar);
	toolBar->addSeparator();
	loadAction->addTo(toolBar);
	saveAction->addTo(toolBar);
	toolBar->addSeparator();
	singleViewAction->addTo(toolBar);
	splitViewAction->addTo(toolBar);
	fullViewAction->addTo(toolBar);

	// create config menu
	QPopupMenu* config = new QPopupMenu(this);
	menu->insertItem("&File", config);
	loadAction->addTo(config);
	saveAction->addTo(config);
	saveAsAction->addTo(config);
	config->insertSeparator();
	quitAction->addTo(config);

	// create options menu
	QPopupMenu* optionMenu = new QPopupMenu(this);
	menu->insertItem("&Option", optionMenu);
	showAllAction->addTo(optionMenu);
	showNameAction->addTo(optionMenu);
	showRangeAction->addTo(optionMenu);
	showDataAction->addTo(optionMenu);

	connect(configList, SIGNAL(selectionChanged(QListViewItem*)),
		SLOT(setHelp(QListViewItem*)));

	showFullView();
}

/*
 * display a new help entry as soon as a new menu entry is selected
 */
void ConfigView::setHelp(QListViewItem* item)
{
	struct symbol* sym;
	struct menu* menu;

	configList->lineEdit->hide();
	if (item) {
		menu = ((ConfigItem*)item)->menu;

		sym = menu->sym;
		if (sym) {
			helpText->setText(sym->help);
			return;
		}
	}
	helpText->setText(NULL);
}

void ConfigView::loadConfig(void)
{
	QString s = QFileDialog::getOpenFileName(conf_filename, NULL, this);
	if (s.isNull())
		return;
	if (conf_read(s.latin1()))
		QMessageBox::information(this, "qconf", "Unable to load configuration!");
}

void ConfigView::saveConfig(void)
{
	if (conf_write(NULL))
		QMessageBox::information(this, "qconf", "Unable to save configuration!");
}

void ConfigView::saveConfigAs(void)
{
	QString s = QFileDialog::getSaveFileName(conf_filename, NULL, this);
	if (s.isNull())
		return;
	if (conf_write(s.latin1()))
		QMessageBox::information(this, "qconf", "Unable to save configuration!");
}

void ConfigView::showSingleView(void)
{
	menuList->hide();
	menuList->setRootMenu(0);
	configList->showFull = false;
	if (configList->rootEntry == &rootmenu)
		configList->updateListAll();
	else
		configList->setRootMenu(&rootmenu);
}

void ConfigView::showSplitView(void)
{
	configList->showFull = true;
	if (configList->rootEntry)
		configList->setRootMenu(0);
	else
		configList->updateListAll();
	configApp->processEvents();
	menuList->showFull = false;
	menuList->setRootMenu(&rootmenu);
	menuList->show();
}

void ConfigView::showFullView(void)
{
	menuList->hide();
	menuList->setRootMenu(0);
	configList->showFull = true;
	if (configList->rootEntry == &rootmenu)
		configList->updateListAll();
	else
		configList->setRootMenu(&rootmenu);
}

void ConfigView::setShowAll(bool b)
{
	if (configList->showAll == b)
		return;
	configList->showAll = b;
	configList->updateListAll();
}

void ConfigView::setShowName(bool b)
{
	if (configList->showName == b)
		return;
	configList->showName = b;
	configList->reinit();
}

void ConfigView::setShowRange(bool b)
{
	if (configList->showRange == b)
		return;
	configList->showRange = b;
	configList->reinit();
}

void ConfigView::setShowData(bool b)
{
	if (configList->showData == b)
		return;
	configList->showData = b;
	configList->reinit();
}

/*
 * ask for saving configuration before quitting
 * TODO ask only when something changed
 */
void ConfigView::closeEvent(QCloseEvent* e)
{
	QMessageBox mb("qconf", "Save configuration?", QMessageBox::Warning,
			QMessageBox::Yes | QMessageBox::Default, QMessageBox::No, QMessageBox::Cancel | QMessageBox::Escape);
	switch (mb.exec()) {
	case QMessageBox::Yes:
		conf_write(NULL);
	case QMessageBox::No:
		e->accept();
		break;
	case QMessageBox::Cancel:
		e->ignore();
		break;
	}
}

int main(int ac, char** av)
{
	ConfigView* v;
	const char *name;

	configApp = new QApplication(ac, av);
	if (ac > 1 && av[1][0] == '-') {
		switch (av[1][1]) {
		case 'a':
			//showAll = 1;
			break;
		case 'h':
		case '?':
			printf("%s <config>\n", av[0]);
			exit(0);
		}
		name = av[2];
	} else
		name = av[1];
	conf_parse(name);
	conf_read(NULL);
	//zconfdump(stdout);
	v = new ConfigView();

	//zconfdump(stdout);
	v->show();
	configApp->connect(configApp, SIGNAL(lastWindowClosed()), SLOT(quit()));
	configApp->exec();
	return 0;
}
